@@ -1,25 +1,26 @@
 /* calc_sums.c - crc calculating and printing functions */
 
-#include "common_func.h" /* should be included before the C library files */
+#include "platform.h" /* unlink() on unix */
 #include <stdio.h>
 #include <string.h>
 #include <stdlib.h> /* free() */
-#include <unistd.h> /* read() */
-#include <fcntl.h>  /* open() */
-#include <time.h>   /* localtime(), time() */
-#include <sys/stat.h> /* stat() */
 #include <errno.h>
 #include <assert.h>
-
-#include "librhash/rhash.h"
-#include "librhash/rhash_torrent.h"
-#include "parse_cmdline.h"
-#include "rhash_main.h"
-#include "hash_print.h"
-#include "output.h"
-#include "win_utils.h"
+#ifdef _WIN32
+# include <fcntl.h>  /* _O_BINARY */
+# include <io.h>
+#endif
 
 #include "calc_sums.h"
+#include "common_func.h"
+#include "file.h"
+#include "hash_print.h"
+#include "output.h"
+#include "parse_cmdline.h"
+#include "rhash_main.h"
+#include "win_utils.h"
+#include "librhash/rhash.h"
+#include "librhash/rhash_torrent.h"
 
 /**
  * Initialize BTIH hash function. Unlike other algorithms BTIH
@@ -123,10 +124,8 @@ static int calc_sums(struct file_info *info)
 
 		/* skip without reporting an error the files
 		 * opened exclusively by another process */
-		fd = rsh_fopen_bin(info->full_path, "rb");
-		if (!fd) {
-			return -1;
-		}
+		fd = file_fopen(info->file, FOpenRead | FOpenBin);
+		if (!fd) return -1;
 	}
 
 	re_init_rhash_context(info);
@@ -218,9 +217,10 @@ const char* file_info_get_utf8_print_path(struct file_info* info)
  * @param crc32 pointer to integer to receive parsed hash sum.
  * @return non zero if crc32 was found, zero otherwise.
  */
-static int find_embedded_crc32(const char* filepath, unsigned* crc32_be)
+static int find_embedded_crc32(const char* filepath, unsigned* crc32)
 {
 	const char* e = filepath + strlen(filepath) - 10;
+	unsigned char raw[4];
 
 	/* search for the sum enclosed in brackets */
 	for (; e >= filepath && !IS_PATH_SEPARATOR(*e); e--) {
@@ -228,7 +228,9 @@ static int find_embedded_crc32(const char* filepath, unsigned* crc32_be)
 			const char *p = e + 8;
 			for (; p > e && IS_HEX(*p); p--);
 			if (p == e) {
-				rhash_hex_to_byte(e + 1, (char unsigned*)crc32_be, 8);
+				rhash_hex_to_byte(e + 1, raw, 8);
+				*crc32 = ((unsigned)raw[0] << 24) | ((unsigned)raw[1] << 16) |
+					((unsigned)raw[2] << 8) | (unsigned)raw[3];
 				return 1;
 			}
 			e -= 9;
@@ -251,18 +253,13 @@ int rename_file_by_embeding_crc32(struct file_info *info)
 	const char* c = p - 1;
 	char* new_path;
 	char* insertion_point;
-	unsigned crc32_be;
+	unsigned crc32;
 	assert((info->rctx->hash_id & RHASH_CRC32) != 0);
 
-	/* check if the filename contains a CRC32 hash sum */
-	if (find_embedded_crc32(info->print_path, &crc32_be)) {
-		unsigned char* c =
-			(unsigned char*)rhash_get_context_ptr(info->rctx, RHASH_CRC32);
-		unsigned actual_crc32 = ((unsigned)c[0] << 24) |
-			((unsigned)c[1] << 16) | ((unsigned)c[2] << 8) | (unsigned)c[3];
-
+	/* check if the filename contains a CRC32 sum */
+	if (find_embedded_crc32(info->print_path, &crc32)) {
 		/* compare with calculated CRC32 */
-		if (crc32_be != actual_crc32) {
+		if (crc32 != get_crc32(info->rctx)) {
 			char crc32_str[9];
 			rhash_print(crc32_str, info->rctx, RHASH_CRC32, RHPR_UPPERCASE);
 			/* TRANSLATORS: sample filename with embedded CRC32: file_[A1B2C3D4].mkv */
@@ -317,6 +314,7 @@ int rename_file_by_embeding_crc32(struct file_info *info)
  */
 int save_torrent_to(const char* path, rhash_context* rctx)
 {
+	file_t file;
 	FILE* fd;
 	int res = 0;
 
@@ -336,7 +334,8 @@ int save_torrent_to(const char* path, rhash_context* rctx)
 	}
 
 	/* write the torrent file */
-	fd = rsh_fopen_bin(path, "wb");
+	file_init(&file, path, FILE_OPT_DONT_FREE_PATH);
+	fd = file_fopen(&file, FOpenWrite | FOpenBin);
 	if (fd && text->length == fwrite(text->str, 1, text->length, fd) &&
 		!ferror(fd) && !fflush(fd))
 	{
@@ -346,6 +345,7 @@ int save_torrent_to(const char* path, rhash_context* rctx)
 		res = -1;
 	}
 	if (fd) fclose(fd);
+	file_cleanup(&file);
 	return res;
 }
 
@@ -399,7 +399,7 @@ int calculate_and_print_sums(FILE* out, file_t* file, const char *print_path)
 		/* calculate sums */
 		if (calc_sums(&info) < 0) {
 			/* print i/o error */
-			log_file_error(file->path);
+			log_file_t_error(file);
 			res = -1;
 		}
 		if (rhash_data.interrupted) {
@@ -433,12 +433,10 @@ int calculate_and_print_sums(FILE* out, file_t* file, const char *print_path)
 	if (rhash_data.print_list && res >= 0) {
 		if (!opt.bt_batch_file) {
 			print_line(out, rhash_data.print_list, &info);
-			fflush(out);
 
 			/* print calculated line to stderr or log-file if verbose */
 			if ((opt.mode & MODE_UPDATE) && (opt.flags & OPT_VERBOSE)) {
 				print_line(rhash_data.log, rhash_data.print_list, &info);
-				fflush(rhash_data.log);
 			}
 		}
 
@@ -479,7 +477,7 @@ static int verify_sums(struct file_info *info)
 	}
 
 	if ((opt.flags & OPT_EMBED_CRC) && find_embedded_crc32(
-		info->print_path, &info->hc.embedded_crc32_be)) {
+		info->print_path, &info->hc.embedded_crc32)) {
 			info->hc.flags |= HC_HAS_EMBCRC32;
 			assert(info->hc.hash_mask & RHASH_CRC32);
 	}
@@ -518,8 +516,8 @@ int check_hash_file(file_t* file, int chdir)
 
 	/* process --check-embedded option */
 	if (opt.mode & MODE_CHECK_EMBEDDED) {
-		unsigned crc32_be;
-		if (find_embedded_crc32(hash_file_path, &crc32_be)) {
+		unsigned crc32;
+		if (find_embedded_crc32(hash_file_path, &crc32)) {
 			/* initialize file_info structure */
 			memset(&info, 0, sizeof(info));
 			info.full_path = rsh_strdup(hash_file_path);
@@ -527,7 +525,7 @@ int check_hash_file(file_t* file, int chdir)
 			file_info_set_print_path(&info, info.full_path);
 			info.sums_flags = info.hc.hash_mask = RHASH_CRC32;
 			info.hc.flags = HC_HAS_EMBCRC32;
-			info.hc.embedded_crc32_be = crc32_be;
+			info.hc.embedded_crc32 = crc32;
 
 			res = verify_sums(&info);
 			fflush(rhash_data.out);
@@ -553,14 +551,14 @@ int check_hash_file(file_t* file, int chdir)
 	if (file->mode & FILE_IFSTDIN) {
 		fd = stdin;
 		hash_file_path = "<stdin>";
-	} else if ( !(fd = rsh_fopen_bin(hash_file_path, "rb") )) {
+	} else if ( !(fd = file_fopen(file, FOpenRead | FOpenBin) )) {
 		log_file_error(hash_file_path);
 		return -1;
 	}
 
 	pos = strlen(hash_file_path)+16;
 	ralign = str_set(buf, '-', (pos < 80 ? 80 - (int)pos : 2));
-	fprintf(rhash_data.out, _("\n--( Verifying %s )%s\n"), hash_file_path, ralign);
+	rsh_fprintf(rhash_data.out, _("\n--( Verifying %s )%s\n"), hash_file_path, ralign);
 	fflush(rhash_data.out);
 	rsh_timer_start(&timer);
 
@@ -572,8 +570,7 @@ int check_hash_file(file_t* file, int chdir)
 	} else pos = 0;
 
 	/* read crc file line by line */
-	for (line_num = 0; fgets(buf, 2048, fd); line_num++)
-	{
+	for (line_num = 0; fgets(buf, 2048, fd); line_num++) {
 		char* line = buf;
 		char* path_without_ext = NULL;
 
@@ -627,7 +624,7 @@ int check_hash_file(file_t* file, int chdir)
 			}
 			memset(&file_to_check, 0, sizeof(file_t));
 			file_to_check.path = info.full_path;
-			file_stat(&file_to_check);
+			file_stat(&file_to_check, 0);
 			info.file = &file_to_check;
 
 			/* verify hash sums of the file */
@@ -651,7 +648,7 @@ int check_hash_file(file_t* file, int chdir)
 	}
 	time = rsh_timer_stop(&timer);
 
-	fprintf(rhash_data.out, "%s\n", str_set(buf, '-', 80));
+	rsh_fprintf(rhash_data.out, "%s\n", str_set(buf, '-', 80));
 	print_check_stats();
 
 	if (rhash_data.processed != rhash_data.ok) rhash_data.error_flag = 1;
@@ -761,7 +758,7 @@ void run_benchmark(unsigned hash_id, unsigned flags)
 		total_time += time;
 
 		if ((flags & BENCHMARK_RAW) == 0 && !rhash_data.interrupted) {
-			fprintf(rhash_data.out, "%s %u MiB calculated in %.3f sec, %.3f MBps\n", hash_name, (unsigned)sz_mb, time, (double)sz_mb / time);
+			rsh_fprintf(rhash_data.out, "%s %u MiB calculated in %.3f sec, %.3f MBps\n", hash_name, (unsigned)sz_mb, time, (double)sz_mb / time);
 			fflush(rhash_data.out);
 		}
 	}
@@ -798,20 +795,20 @@ void run_benchmark(unsigned hash_id, unsigned flags)
 
 	if (flags & BENCHMARK_RAW) {
 		/* output result in a "raw" machine-readable format */
-		fprintf(rhash_data.out, "%s\t%u\t%.3f\t%.3f", hash_name, ((unsigned)sz_mb * rounds), total_time, (double)(sz_mb * rounds) / total_time);
+		rsh_fprintf(rhash_data.out, "%s\t%u\t%.3f\t%.3f", hash_name, ((unsigned)sz_mb * rounds), total_time, (double)(sz_mb * rounds) / total_time);
 #if defined(HAVE_TSC)
 		if (flags & BENCHMARK_CPB) {
-			fprintf(rhash_data.out, "\t%.2f", cpb);
+			rsh_fprintf(rhash_data.out, "\t%.2f", cpb);
 		}
 #endif /* HAVE_TSC */
-		fprintf(rhash_data.out, "\n");
+		rsh_fprintf(rhash_data.out, "\n");
 	} else {
-		fprintf(rhash_data.out, "%s %u MiB total in %.3f sec, %.3f MBps", hash_name, ((unsigned)sz_mb * rounds), total_time, (double)(sz_mb * rounds) / total_time);
+		rsh_fprintf(rhash_data.out, "%s %u MiB total in %.3f sec, %.3f MBps", hash_name, ((unsigned)sz_mb * rounds), total_time, (double)(sz_mb * rounds) / total_time);
 #if defined(HAVE_TSC)
 		if (flags & BENCHMARK_CPB) {
-			fprintf(rhash_data.out, ", CPB=%.2f", cpb);
+			rsh_fprintf(rhash_data.out, ", CPB=%.2f", cpb);
 		}
 #endif /* HAVE_TSC */
-		fprintf(rhash_data.out, "\n");
+		rsh_fprintf(rhash_data.out, "\n");
 	}
 }
