@@ -1,25 +1,21 @@
-/* common_func.c */
+/* common_func.c - functions used almost everywhere */
 
-#include "common_func.h" /* should be included before the C library files */
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
-#include <ctype.h>
 #include <assert.h>
-#include <stdint.h>
-#include <time.h>
-#include <stdarg.h>
-#include <wchar.h>
+#include <ctype.h>
 #include <errno.h>
-#include <sys/stat.h>
+#include <stdarg.h>
+#include <stdlib.h>
+#include <string.h>
+#include <wchar.h>
 
-#ifdef _WIN32
-#include <windows.h>
+#if defined( _WIN32) || defined(__CYGWIN__)
+# include <windows.h>
 #endif
 
-#include "win_utils.h"
+#include "common_func.h"
 #include "parse_cmdline.h"
 #include "version.h"
+#include "win_utils.h"
 
 /*=========================================================================
  * String functions
@@ -233,109 +229,6 @@ const char* get_bt_program_name(void)
 	return bt_program_name;
 }
 
-/*=========================================================================
- * Path functions
- *=========================================================================*/
-
-/**
- * Return filename without path.
- *
- * @param path file path
- * @return filename
- */
-const char* get_basename(const char* path)
-{
-	const char *p = path + strlen(path) - 1;
-	for (; p >= path && !IS_PATH_SEPARATOR(*p); p--);
-	return (p+1);
-}
-
-/**
- * Return allocated buffer with the directory part of the path.
- * The buffer must be freed by calling free().
- *
- * @param path file path
- * @return directory
- */
-char* get_dirname(const char* path)
-{
-	const char *p = path + strlen(path) - 1;
-	char *res;
-	for (; p > path && !IS_PATH_SEPARATOR(*p); p--);
-	if ((p - path) > 1) {
-		res = (char*)rsh_malloc(p-path+1);
-		memcpy(res, path, p-path);
-		res[p-path] = 0;
-		return res;
-	} else {
-		return rsh_strdup(".");
-	}
-}
-
-/**
- * Assemble a filepath from its directory and filename.
- *
- * @param dir_path directory path
- * @param filename file name
- * @return filepath
- */
-char* make_path(const char* dir_path, const char* filename)
-{
-	char* buf;
-	size_t len;
-	assert(dir_path);
-	assert(filename);
-
-	/* remove leading path separators from filename */
-	while (IS_PATH_SEPARATOR(*filename)) filename++;
-
-	if (dir_path[0] == '.' && dir_path[1] == 0) {
-		/* do not extend filename for dir_path="." */
-		return rsh_strdup(filename);
-	}
-
-	/* copy directory path */
-	len = strlen(dir_path);
-	buf = (char*)rsh_malloc(len + strlen(filename) + 2);
-	strcpy(buf, dir_path);
-
-	/* separate directory from filename */
-	if (len > 0 && !IS_PATH_SEPARATOR(buf[len-1])) {
-		buf[len++] = SYS_PATH_SEPARATOR;
-	}
-
-	/* append filename */
-	strcpy(buf+len, filename);
-	return buf;
-}
-
-#define IS_ANY_SLASH(c) ((c) == RSH_T('/') || (c) == RSH_T('\\'))
-
-/**
- * Compare paths.
- *
- * @param a the first path
- * @param b the second path
- */
-int are_paths_equal(const rsh_tchar* a, const rsh_tchar* b)
-{
-	if (!a || !b) return 0;
-	if (a[0] == RSH_T('.') && IS_ANY_SLASH(a[1])) a += 2;
-	if (b[0] == RSH_T('.') && IS_ANY_SLASH(b[1])) b += 2;
-	
-	for (; *a; ++a, ++b)
-	{
-		if (*a != *b && (!IS_ANY_SLASH(*b) || !IS_ANY_SLASH(*a)))
-		{
-			/* paths are different */
-			return 0;
-		}
-		
-	}
-	/* check if both paths terminated */
-	return (*a == *b);
-}
-
 /*=========================================================================
  * Timer functions
  *=========================================================================*/
@@ -350,7 +243,7 @@ int are_paths_equal(const rsh_tchar* a, const rsh_tchar* b)
  */
 static double rsh_fsec(timedelta_t* timer)
 {
-#ifdef _WIN32
+#if defined( _WIN32) || defined(__CYGWIN__)
 	LARGE_INTEGER freq;
 	QueryPerformanceFrequency(&freq);
 	return (double)*timer / freq.QuadPart;
@@ -359,8 +252,7 @@ static double rsh_fsec(timedelta_t* timer)
 #endif
 }
 
-#ifdef _WIN32
-#include <windows.h>
+#if defined( _WIN32) || defined(__CYGWIN__)
 #define get_timedelta(delta) QueryPerformanceCounter((LARGE_INTEGER*)delta)
 #else
 #define get_timedelta(delta) gettimeofday(delta, NULL)
@@ -375,7 +267,7 @@ double rsh_timer_stop(timedelta_t* timer)
 {
 	timedelta_t end;
 	get_timedelta(&end);
-#ifdef _WIN32
+#if defined( _WIN32) || defined(__CYGWIN__)
 	*timer = end - *timer;
 #else
 	timer->tv_sec  = end.tv_sec  - timer->tv_sec - (end.tv_usec >= timer->tv_usec ? 0 : 1);
@@ -386,7 +278,7 @@ double rsh_timer_stop(timedelta_t* timer)
 
 unsigned rhash_get_ticks(void)
 {
-#ifdef _WIN32
+#if defined( _WIN32) || defined(__CYGWIN__)
 	return GetTickCount();
 #else
 	struct timeval tv;
@@ -395,179 +287,54 @@ unsigned rhash_get_ticks(void)
 #endif
 }
 
-/*=========================================================================
- * File functions
- *=========================================================================*/
-
-void file_init(file_t* file, const char* path, int reuse_path)
-{
-	memset(file, 0, sizeof(*file));
-	if (reuse_path)
-	{
-		file->path = (char*)path;
-		file->mode = FILE_OPT_DONT_FREE_PATH;
-	} else {
-		file->path = rsh_strdup(path);
-	}
-}
-
-void file_cleanup(file_t* file)
-{
-	if ((file->mode & FILE_OPT_DONT_FREE_PATH) == 0)
-	{
-		free(file->path);
-	}
-	file->path = NULL;
-
-#ifdef _WIN32
-	free(file->wpath);
-	file->wpath = NULL;
-#endif /* _WIN32 */
-
-	file->mtime = file->size = 0;
-	file->mode = 0;
-}
-
-#ifdef _WIN32
-/**
- * Fill file information in the file_t structure.
- *
- * @param file the file information
- * @return 0 on success, -1 on error
- */
-int file_statw(file_t* file)
-{
-	WIN32_FILE_ATTRIBUTE_DATA data;
-
-	/* read file attributes */
-	if (GetFileAttributesExW(file->wpath, GetFileExInfoStandard, &data)) {
-		uint64_t u;
-		file->size  = (((uint64_t)data.nFileSizeHigh) << 32) + data.nFileSizeLow;
-		file->mode &= FILE_OPT_DONT_FREE_PATH;
-		file->mode |= (data.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY ? FILE_IFDIR : FILE_IFREG);
-
-		/* the number of 100-nanosecond intervals since January 1, 1601 */
-		u = (((uint64_t)data.ftLastWriteTime.dwHighDateTime) << 32) + data.ftLastWriteTime.dwLowDateTime;
-		/* convert to second and subtract the epoch difference */
-		file->mtime = u / 10000000 - 11644473600LL;
-		return 0;
-	}
-	set_errno_from_last_file_error();
-	return -1;
-}
-#endif
-
-/**
- * Fill file information in the file_t structure.
- *
- * @param file the file information
- * @param use_lstat nonzero if lstat() shall be used
- * @return 0 on success, -1 on error
- */
-int file_stat2(file_t* file, int use_lstat)
-{
-#ifdef _WIN32
-	int i;
-	(void)use_lstat; /* ignore on windows */
-
-	file->mtime = 0;
-	if (file->wpath) {
-		free(file->wpath);
-		file->wpath = NULL;
-	}
-
-	for (i = 0; i < 2; i++) {
-		file->wpath = c2w(file->path, i);
-		if (file->wpath == NULL) continue;
-
-		/* return on success */
-		if (file_statw(file) == 0) return 0;
-
-		free(file->wpath);
-		file->wpath = NULL;
-	}
-	/* NB: usually errno is set by the rsh_file_statw() call */
-	if (!errno) errno = EINVAL;
-	return -1;
-#else
-	struct stat st;
-	int res = 0;
-	file->mode &= FILE_OPT_DONT_FREE_PATH;
-
-	do {
-		if (use_lstat) {
-			/* check for symlink */
-			if (lstat(file->path, &st) < 0) return -1;
-			if (!S_ISLNK(st.st_mode)) break;
-
-			/* it's a symlink */
-			file->mode |= FILE_IFLNK;
-		}
-
-		res = stat(file->path, &st);
-	} while (0);
-
-	file->size  = st.st_size;
-	file->mtime = st.st_mtime;
-
-	if (S_ISDIR(st.st_mode)) {
-		file->mode |= FILE_IFDIR;
-	} else if (S_ISREG(st.st_mode)) {
-		/* it's a regular file or a symlink pointing to a regular file */
-		file->mode |= FILE_IFREG;
-	}
-
-	return res;
-#endif /* _WIN32 */
-}
-
-/**
- * Read the file information like its type, size and modification date.
- *
- * @param file the file information
- * @return 0 on success, -1 on error
- */
-int file_stat(file_t* file)
-{
-	return file_stat2(file, 0);
-}
-
-int is_regular_file(const char* path)
-{
-	int is_regular = 0;
-	file_t file;
-
-	file_init(&file, path, 1);
-	if (file_stat(&file) >= 0) {
-		is_regular = FILE_ISREG(&file);
-	}
-	file_cleanup(&file);
-	return is_regular;
-}
-
-int if_file_exists(const char* path)
-{
-	int exists;
-	file_t file;
-
-	file_init(&file, path, 1);
-	exists = (file_stat(&file) >= 0);
-	file_cleanup(&file);
-	return exists;
-}
-
 /*=========================================================================
  * Custom program exit function
  *=========================================================================*/
 
+struct rhash_exit_handlers_t
+{
+	unsigned handlers_count;
+	exit_handler_t handlers[4];
+} rhash_exit_handlers = { 0, { 0 } };
+
 /**
-* Exit the program, with restoring console state.
+* Install a handler to be called on program exit.
+*
+* @param handler the hadler to add
+*/
+void rsh_install_exit_handler(exit_handler_t handler)
+{
+	if (rhash_exit_handlers.handlers_count >= (sizeof(rhash_exit_handlers.handlers) / sizeof(rhash_exit_handlers.handlers[0])))
+	{
+		assert(!"to many handlers");
+		rsh_exit(2);
+	}
+	rhash_exit_handlers.handlers[rhash_exit_handlers.handlers_count] = handler;
+	rhash_exit_handlers.handlers_count++;
+}
+
+/**
+* Remove the last installed exit handler.
+*/
+void rsh_remove_exit_handler(void)
+{
+	if (rhash_exit_handlers.handlers_count == 0)
+	{
+		assert(rhash_exit_handlers.handlers_count > 0 && "no handlers installed");
+		rsh_exit(2);
+	}
+	rhash_exit_handlers.handlers_count--;
+}
+
+/**
+* Call all installed exit handlers, starting from the latest one, and exit the program.
 *
 * @param code the program exit code
 */
-void rhash_exit(int code)
+void rsh_exit(int code)
 {
-	IF_WINDOWS(restore_console());
+	while (rhash_exit_handlers.handlers_count > 0)
+		rhash_exit_handlers.handlers[--rhash_exit_handlers.handlers_count]();
 	exit(code);
 }
 
@@ -578,7 +345,6 @@ void rhash_exit(int code)
 static void report_error_default(const char* srcfile, int srcline,
 	const char* format, ...);
 
-void (*rsh_exit)(int code) = exit;
 void (*rsh_report_error)(const char* srcfile, int srcline,
 	const char* format, ...) = report_error_default;
 
@@ -592,9 +358,9 @@ void (*rsh_report_error)(const char* srcfile, int srcline,
 static void report_error_default(const char* srcfile, int srcline, const char* format, ...)
 {
 	va_list ap;
-	fprintf(stderr, "RHash: error at %s:%u: ", srcfile, srcline);
+	rsh_fprintf(stderr, "RHash: error at %s:%u: ", srcfile, srcline);
 	va_start(ap, format);
-	vfprintf(stderr, format, ap); /* report the error to stderr */
+	rsh_vfprintf(stderr, format, ap); /* report the error to stderr */
 	va_end(ap);
 }
 
