@@ -1,12 +1,9 @@
-/* find_file.h - declaration of find_file function.
- *
- * find_file function searches through a directory tree calling a call_back on
- * each file.
- */
+/* find_file.h - functions for recursive scan of directories. */
 #ifndef FIND_FILE_H
 #define FIND_FILE_H
 
 #include "common_func.h"
+#include "file.h"
 
 #ifdef __cplusplus
 extern "C" {
@@ -32,8 +29,9 @@ typedef struct file_search_data
 	int errors_count;
 } file_search_data;
 
-file_search_data* create_file_search_data(rsh_tchar** paths, size_t count, int max_depth);
-void destroy_file_search_data(file_search_data* data);
+file_search_data* file_search_data_new(void);
+void file_search_add_file(file_search_data* data, tstr_t path, int is_file_list);
+void file_search_data_free(file_search_data* data);
 
 void scan_files(file_search_data* data);
 
