@@ -1,17 +1,21 @@
-/* hash_print.c - functions to output hash sums using printf-like format */
+/* hash_print.c - output hash sums using printf-like format */
 
-#include "common_func.h" /* should be included before the C library files */
 #include <string.h>
 #include <stdlib.h>
-#include <stdio.h>
 #include <ctype.h>
 #include <assert.h>
+#ifdef _WIN32
+# include <fcntl.h>
+# include <io.h>
+#endif /* _WIN32 */
 
-#include "librhash/rhash.h"
+#include "hash_print.h"
 #include "calc_sums.h"
+#include "common_func.h"
+#include "file.h"
 #include "parse_cmdline.h"
 #include "win_utils.h"
-#include "hash_print.h"
+#include "librhash/rhash.h"
 
 /*=========================================================================
 * Formatted output functions and structures
@@ -37,6 +41,7 @@ enum {
 		| PRINT_FLAG_HEX | PRINT_FLAG_BASE32 | PRINT_FLAG_BASE64,
 	PRINT_STR = 0x10000000,
 	PRINT_ZERO,
+	PRINT_NEWLINE,
 	PRINT_FILEPATH,
 	PRINT_BASENAME,
 	PRINT_URLNAME,
@@ -77,6 +82,7 @@ static char parse_escaped_char(const char **pformat)
 {
 	const char* start = *pformat;
 	switch ( *((*pformat)++) ) {
+		case '0': return '\0';
 		case 't': return '\t';
 		case 'r': return '\r';
 		case 'n': return '\n';
@@ -92,7 +98,7 @@ static char parse_escaped_char(const char **pformat)
 					ch = 16 * ch + (**pformat <= '9' ? **pformat & 15 : (**pformat + 9) & 15);
 					(*pformat)++;
 				}
-				if (ch) return ch;
+				return ch;
 			}
 			break;
 		default:
@@ -131,11 +137,16 @@ print_item* parse_print_string(const char* format, unsigned *sum_mask)
 			*(p++) = *(format++);
 
 		if (*format == '\\') {
-			if (*(++format) == '0') {
+			format++;
+			*p = parse_escaped_char(&format);
+			if (*p == '\0') {
 				item = new_print_item(PRINT_ZERO, 0, NULL);
-				format++;
+#ifdef _WIN32
+			} else if (*p == '\n') {
+				item = new_print_item(PRINT_NEWLINE, 0, NULL);
+#endif
 			} else {
-				*p++ = parse_escaped_char(&format);
+				p++;
 				continue;
 			}
 		} else if (*format == '%') {
@@ -335,7 +346,7 @@ static void fprint_ed2k_url(FILE* out, struct file_info *info, int print_type)
 		dst += 32;
 	}
 	strcpy(dst, "|/");
-	fprintf(out, "%s", buf);
+	rsh_fprintf(out, "%s", buf);
 	free(buf);
 }
 
@@ -355,7 +366,7 @@ static void fprintI64(FILE* out, uint64_t filesize, int width, int zero_pad)
 	if (len < width && zero_pad) {
 		memset(buf, '0', width-len);
 	}
-	fprintf(out, "%s", buf);
+	rsh_fprintf(out, "%s", buf);
 	free(buf);
 }
 
@@ -375,7 +386,7 @@ static void print_time(FILE *out, time_t time)
 		t->tm_hour = t->tm_min = t->tm_sec =
 			t->tm_year = t->tm_mon = t->tm_mday = 0;
 	}
-	fprintf(out, "%02u:%02u.%02u %4u-%02u-%02u", t->tm_hour, t->tm_min,
+	rsh_fprintf(out, "%02u:%02u.%02u %4u-%02u-%02u", t->tm_hour, t->tm_min,
 		t->tm_sec, (1900 + t->tm_year), t->tm_mon + 1, t->tm_mday);
 }
 
@@ -402,6 +413,11 @@ void print_line(FILE* out, print_item* list, struct file_info *info)
 	const char* basename = get_basename(info->print_path), *tmp;
 	char *url = NULL, *ed2k_url = NULL;
 	char buffer[130];
+#ifdef _WIN32
+	/* switch to binary mode to correctly output binary hashes */
+	int out_fd = _fileno(out);
+	int old_mode = (out_fd > 0 && !isatty(out_fd) ? _setmode(out_fd, _O_BINARY) : -1);
+#endif
 
 	for (; list; list = list->next) {
 		int print_type = list->flags & ~(PRINT_FLAGS_ALL);
@@ -422,23 +438,28 @@ void print_line(FILE* out, print_item* list, struct file_info *info)
 			assert(len < sizeof(buffer));
 
 			/* output the hash, exit on fail */
-			if (fwrite(buffer, 1, len, out) != len) break;
+			rsh_fwrite(buffer, 1, len, out);
 			continue;
 		}
 
 		/* output other special items: filepath, URL-encoded filename etc. */
 		switch (print_type) {
 			case PRINT_STR:
-				fprintf(out, "%s", list->data);
+				rsh_fprintf(out, "%s", list->data);
 				break;
 			case PRINT_ZERO: /* the '\0' character */
-				fprintf(out, "%c", 0);
+				rsh_fprintf(out, "%c", 0);
 				break;
+#ifdef _WIN32
+			case PRINT_NEWLINE:
+				rsh_fprintf(out, "%s", "\r\n");
+				break;
+#endif
 			case PRINT_FILEPATH:
-				fprintf(out, "%s", info->print_path);
+				rsh_fprintf(out, "%s", info->print_path);
 				break;
 			case PRINT_BASENAME: /* the filename without directory */
-				fprintf(out, "%s", basename);
+				rsh_fprintf(out, "%s", basename);
 				break;
 			case PRINT_URLNAME: /* URL-encoded filename */
 				if (!url) {
@@ -446,7 +467,7 @@ void print_line(FILE* out, print_item* list, struct file_info *info)
 					url = (char*)rsh_malloc(urlencode(NULL, tmp) + 1);
 					urlencode(url, tmp);
 				}
-				fprintf(out, "%s", url);
+				rsh_fprintf(out, "%s", url);
 				break;
 			case PRINT_MTIME: /* the last-modified tine of the filename */
 				print_time64(out, info->file->mtime);
@@ -461,6 +482,11 @@ void print_line(FILE* out, print_item* list, struct file_info *info)
 	}
 	free(url);
 	free(ed2k_url);
+	fflush(out);
+#ifdef _WIN32
+	if (old_mode >= 0)
+		_setmode(out_fd, old_mode);
+#endif
 }
 
 /**
@@ -654,9 +680,9 @@ int print_sfv_header_line(FILE* out, file_t* file, const char* printpath)
 	if (printpath[0] == '.' && IS_PATH_SEPARATOR(printpath[1])) printpath += 2;
 
 	sprintI64(buf, file->size, 12);
-	fprintf(out, "; %s  ", buf);
+	rsh_fprintf(out, "; %s  ", buf);
 	print_time64(out, file->mtime);
-	fprintf(out, " %s\n", printpath);
+	rsh_fprintf(out, " %s\n", printpath);
 	return 0;
 }
 
@@ -671,9 +697,9 @@ void print_sfv_banner(FILE* out)
 	time_t cur_time = time(NULL);
 	struct tm *t = localtime(&cur_time);
 	if (t) {
-		fprintf(out, _("; Generated by %s v%s on %4u-%02u-%02u at %02u:%02u.%02u\n"),
+		rsh_fprintf(out, _("; Generated by %s v%s on %4u-%02u-%02u at %02u:%02u.%02u\n"),
 			PROGRAM_NAME, get_version_string(),
 			(1900 + t->tm_year), t->tm_mon + 1, t->tm_mday, t->tm_hour, t->tm_min, t->tm_sec);
-		fprintf(out, _("; Written by Aleksey (Akademgorodok) - http://rhash.sourceforge.net/\n;\n"));
+		rsh_fprintf(out, _("; Written by Kravchenko Aleksey (Akademgorodok) - http://rhash.sf.net/\n;\n"));
 	}
 }
