@@ -1,24 +1,23 @@
 /* parse_cmdline.c - parsing of command line options */
 
-#include "common_func.h" /* should be included before the C library files */
 #include <assert.h>
 #include <locale.h>
 #include <stdlib.h>
 #include <stdio.h>
 #include <string.h>
-#include <sys/stat.h> /* stat() */
 #ifdef _WIN32
-#include <windows.h> /* for SetFileApisToOEM(), CharToOem() */
+# include <windows.h> /* for SetFileApisToOEM(), CharToOem() */
 #endif
 
-#include "librhash/rhash.h"
-#include "win_utils.h"
+#include "parse_cmdline.h"
+#include "common_func.h"
 #include "file_mask.h"
 #include "find_file.h"
 #include "hash_print.h"
 #include "output.h"
 #include "rhash_main.h"
-#include "parse_cmdline.h"
+#include "win_utils.h"
+#include "librhash/rhash.h"
 
 typedef struct options_t options_t;
 struct options_t conf_opt; /* config file parsed options */
@@ -85,11 +84,14 @@ static void print_help(void)
 	print_help_line("  -B, --benchmark  ", _("Benchmark selected algorithm.\n"));
 	print_help_line("  -v, --verbose ", _("Be verbose.\n"));
 	print_help_line("  -r, --recursive  ", _("Process directories recursively.\n"));
+	print_help_line("      --file-list=<file> ", _("Process a list of files.\n"));
 	print_help_line("      --skip-ok ", _("Don't print OK messages for successfully verified files.\n"));
 	print_help_line("  -i, --ignore-case  ", _("Ignore case of filenames when updating hash files.\n"));
 	print_help_line("      --percents   ", _("Show percents, while calculating or checking hashes.\n"));
 	print_help_line("      --speed   ", _("Output per-file and total processing speed.\n"));
 	print_help_line("      --maxdepth=<n> ", _("Descend at most <n> levels of directories.\n"));
+	if (rhash_is_openssl_supported())
+		print_help_line("      --openssl=<list> ", _("List hash functions to be calculated using OpenSSL.\n"));
 	print_help_line("  -o, --output=<file> ", _("File to output calculation or checking results.\n"));
 	print_help_line("  -l, --log=<file>    ", _("File to log errors and verbose information.\n"));
 	print_help_line("      --sfv     ", _("Print hash sums, using SFV format (default).\n"));
@@ -125,6 +127,21 @@ enum file_suffix_type {
 	MASK_CRC_ACCEPT
 };
 
+/**
+ * Add a file-list.
+ *
+ * @param o pointer to the options structure to update
+ * @param path the path to the file-list
+ * @param type the type of the option
+ */
+static void add_file_list(options_t *o, tstr_t path, unsigned is_file_list)
+{
+	if (o->search_data) {
+		file_search_add_file(o->search_data, path, is_file_list);
+		opt.has_files = 1;
+	}
+}
+
 /**
  * Process --accept, --exclude and --crc-accept options.
  *
@@ -165,38 +182,37 @@ static void bt_announce(options_t *o, char* announce_url, unsigned unused)
  */
 static void openssl_flags(options_t *o, char* openssl_hashes, unsigned type)
 {
-#if defined(USE_OPENSSL) || defined(OPENSSL_RUNTIME)
-	char *cur, *next;
 	(void)type;
-	o->openssl_mask = 0x80000000; /* turn off using default mask */
+	if (rhash_is_openssl_supported())
+	{
+		rhash_uptr_t openssl_supported_hashes = rhash_get_openssl_supported_mask();
+		char *cur, *next;
+		o->openssl_mask = 0x80000000; /* turn off using default mask */
 
-	/* set the openssl_mask */
-	for (cur = openssl_hashes; cur && *cur; cur = next) {
-		print_hash_info *info = hash_info_table;
-		unsigned bit;
-		size_t length;
-		next = strchr(cur, ',');
-		length = (next != NULL ? (size_t)(next++ - cur) : strlen(cur));
+		/* set the openssl_mask */
+		for (cur = openssl_hashes; cur && *cur; cur = next) {
+			print_hash_info *info = hash_info_table;
+			unsigned bit;
+			size_t length;
+			next = strchr(cur, ',');
+			length = (next != NULL ? (size_t)(next++ - cur) : strlen(cur));
 
-		for (bit = 1; bit <= RHASH_ALL_HASHES; bit = bit << 1, info++) {
-			if ( (bit & RHASH_OPENSSL_SUPPORTED_HASHES) &&
-				memcmp(cur, info->short_name, length) == 0 &&
-				info->short_name[length] == 0) {
-					o->openssl_mask |= bit;
-					break;
+			for (bit = 1; bit <= RHASH_ALL_HASHES; bit = bit << 1, info++) {
+				if ( (bit & openssl_supported_hashes) &&
+					memcmp(cur, info->short_name, length) == 0 &&
+					info->short_name[length] == 0) {
+						o->openssl_mask |= bit;
+						break;
+				}
+			}
+			if (bit > RHASH_ALL_HASHES) {
+				cur[length] = '\0'; /* terminate wrong hash name */
+				log_warning(_("openssl option doesn't support '%s' hash\n"), cur);
 			}
 		}
-		if (bit > RHASH_ALL_HASHES) {
-			cur[length] = '\0'; /* terminate wrong hash name */
-			log_warning(_("openssl option doesn't support '%s' hash\n"), cur);
-		}
 	}
-#else
-	(void)type;
-	(void)openssl_hashes;
-	(void)o;
-	log_warning(_("compiled without openssl support\n"));
-#endif
+	else
+		log_warning(_("compiled without openssl support\n"));
 }
 
 /**
@@ -300,8 +316,9 @@ enum option_type_t
 	F_TOUT = 3 | F_NEED_PARAM | F_OUTPUT_OPT,
 	F_VFNC = 4, /* just call a function */
 	F_PFNC = 5 | F_NEED_PARAM, /* process option parameter by calling a handler */
-	F_UFNC = 6 | F_NEED_PARAM, /* pass UTF-8 encoded parameter to the handler */
-	F_PRNT = 7, /* print a constant C-string and exit */
+	F_TFNC = 6 | F_NEED_PARAM, /* process option parameter by calling a handler */
+	F_UFNC = 7 | F_NEED_PARAM, /* pass UTF-8 encoded parameter to the handler */
+	F_PRNT = 8, /* print a constant C-string and exit */
 };
 
 #define is_param_required(option_type) ((option_type) & F_NEED_PARAM)
@@ -362,9 +379,8 @@ cmdline_opt_t cmdline_opt[] =
 
 	/* other options */
 	{ F_UFLG, 'r', 'R', "recursive", &opt.flags, OPT_RECURSIVE },
-#ifndef _WIN32 /* Unix-only option */
+	{ F_TFNC,   0,   0, "file-list", add_file_list, 1 },
 	{ F_UFLG,   0,   0, "follow",  &opt.flags, OPT_FOLLOW },
-#endif
 	{ F_UFLG, 'v',   0, "verbose", &opt.flags, OPT_VERBOSE },
 	{ F_UFLG,   0,   0, "gost-reverse", &opt.flags, OPT_GOST_REVERSE },
 	{ F_UFLG,   0,   0, "skip-ok", &opt.flags, OPT_SKIP_OK },
@@ -396,6 +412,7 @@ cmdline_opt_t cmdline_opt[] =
 #endif
 	{ 0,0,0,0,0,0 }
 };
+cmdline_opt_t cmdline_file = { F_TFNC, 0, 0, "FILE", add_file_list, 0 };
 
 /**
  * Log a message and exit the program.
@@ -448,7 +465,7 @@ static void apply_option(options_t *opts, parsed_option_t* option)
 		}
 
 #ifdef _WIN32
-		if (option_type == F_TOUT) {
+		if (option_type == F_TOUT || option_type == F_TFNC) {
 			/* leave the value in UTF-16 */
 			value = (char*)rsh_wcsdup((wchar_t*)option->parameter);
 		}
@@ -477,6 +494,7 @@ static void apply_option(options_t *opts, parsed_option_t* option)
 		*(char**)((char*)opts + ((char*)o->ptr - (char*)&opt)) = value;
 		break;
 	case F_PFNC:
+	case F_TFNC:
 	case F_UFNC:
 		/* call option parameter handler */
 		( ( void(*)(options_t *, char*, unsigned) )o->ptr )(opts, value, o->param);
@@ -500,7 +518,11 @@ static void apply_option(options_t *opts, parsed_option_t* option)
  */
 static const char* find_conf_file(void)
 {
-# define CONF_FILE_NAME "rhashrc"
+#ifndef SYSCONFDIR
+# define SYSCONFDIR "/etc"
+#endif
+#define CONFIG_FILENAME "rhashrc"
+
 	char *dir1, *path;
 
 #ifndef _WIN32 /* Linux/Unix part */
@@ -514,7 +536,7 @@ static const char* find_conf_file(void)
 		free(path);
 	}
 	/* then check for global config */
-	path = "/etc/" CONF_FILE_NAME;
+	path = SYSCONFDIR "/" CONFIG_FILENAME;
 	if (is_regular_file(path)) {
 		return (conf_opt.config_file = path);
 	}
@@ -524,7 +546,7 @@ static const char* find_conf_file(void)
 	/* first check for the %APPDATA%\RHash\rhashrc config */
 	if ( (dir1 = getenv("APPDATA")) ) {
 		dir1 = make_path(dir1, "RHash");
-		path = make_path(dir1, CONF_FILE_NAME);
+		path = make_path(dir1, CONFIG_FILENAME);
 		free(dir1);
 		if (is_regular_file(path)) {
 			rsh_vector_add_ptr(opt.mem, path);
@@ -537,7 +559,7 @@ static const char* find_conf_file(void)
 	/* note that %USERPROFILE% is generally not a user home dir */
 	if ( (dir1 = getenv("HOMEDRIVE")) && (path = getenv("HOMEPATH"))) {
 		dir1 = make_path(dir1, path);
-		path = make_path(dir1, CONF_FILE_NAME);
+		path = make_path(dir1, CONFIG_FILENAME);
 		free(dir1);
 		if (is_regular_file(path)) {
 			rsh_vector_add_ptr(opt.mem, path);
@@ -548,7 +570,7 @@ static const char* find_conf_file(void)
 	
 	/* check for ${PROGRAM_DIR}\rhashrc */
 	if (rhash_data.program_dir && (dir1 = w2c(rhash_data.program_dir))) {
-		path = make_path(dir1, CONF_FILE_NAME);
+		path = make_path(dir1, CONFIG_FILENAME);
 		free(dir1);
 		if (is_regular_file(path)) {
 			rsh_vector_add_ptr(opt.mem, path);
@@ -705,13 +727,23 @@ struct parsed_cmd_line_t
 	blocks_vector_t options; /* array of parsed options */
 	int  argc;
 	char **argv;
-	int n_files; /* the number of files */
-	rsh_tchar** files; /* array of files specified by command line */
 #ifdef _WIN32
 	rsh_tchar** warg; /* program arguments in Unicode */
 #endif
 };
 
+/**
+ * Allocate parsed option.
+ *
+ * @param cmd_line the command line to store the parsed option into.
+ * @return allocated parsed option
+ */
+static parsed_option_t* new_option(struct parsed_cmd_line_t* cmd_line)
+{
+	rsh_blocks_vector_add_empty(&cmd_line->options, 16, sizeof(parsed_option_t));
+	return rsh_blocks_vector_get_item(&cmd_line->options, cmd_line->options.size - 1, 16, parsed_option_t);
+}
+
 /**
  * Parse command line arguments.
  *
@@ -720,8 +752,7 @@ struct parsed_cmd_line_t
 static void parse_cmdline_options(struct parsed_cmd_line_t* cmd_line)
 {
 	int argc;
-	int n_files = 0, b_opt_end = 0;
-	rsh_tchar** files;
+	int b_opt_end = 0;
 	rsh_tchar **parg, **end_arg;
 	parsed_option_t *next_opt;
 
@@ -736,7 +767,6 @@ static void parse_cmdline_options(struct parsed_cmd_line_t* cmd_line)
 #endif
 
 	/* allocate array for files */
-	files = (rsh_tchar**)rsh_malloc(argc * sizeof(rsh_tchar*));
 	end_arg = parg + argc;
 
 	/* loop by program arguments */
@@ -744,31 +774,19 @@ static void parse_cmdline_options(struct parsed_cmd_line_t* cmd_line)
 		/* if argument is not an option */
 		if ((*parg)[0] != RSH_T('-') || (*parg)[1] == 0 || b_opt_end) {
 			/* it's a file, note that '-' is interpreted as stdin */
-			files[n_files++] = *parg;
-			continue;
-		}
-
-		assert((*parg)[0] == RSH_T('-') && (*parg)[1] != 0);
-
-		if ((*parg)[1] == L'-' && (*parg)[2] == 0) {
+			next_opt = new_option(cmd_line);
+			next_opt->name = "";
+			next_opt->o = &cmdline_file;
+			next_opt->parameter = *parg;
+		} else if ((*parg)[1] == L'-' && (*parg)[2] == 0) {
 			b_opt_end = 1; /* string "--" means end of options */
 			continue;
-		}
-
-		/* check for "--" */
-		if ((*parg)[1] == RSH_T('-')) {
-			cmdline_opt_t *t;
-
-			/* allocate parsed_option */
-			rsh_blocks_vector_add_empty(&cmd_line->options, 16, sizeof(parsed_option_t));
-			next_opt = rsh_blocks_vector_get_item(&cmd_line->options, cmd_line->options.size - 1, 16, parsed_option_t);
-
-			/* find the long option */
+		} else if ((*parg)[1] == RSH_T('-')) {
+			next_opt = new_option(cmd_line);
 			parse_long_option(next_opt, &parg);
-			t = next_opt->o;
 
 			/* process encoding and -o/-l options early */
-			if (is_output_modifier(t->type)) {
+			if (is_output_modifier(next_opt->o->type)) {
 				apply_option(&opt, next_opt);
 			}
 		} else if ((*parg)[1] != 0) {
@@ -787,10 +805,7 @@ static void parse_cmdline_options(struct parsed_cmd_line_t* cmd_line)
 					fail_on_unknow_option(w2c(ptr));
 				}
 #endif
-				/* allocate parsed_option */
-				rsh_blocks_vector_add_empty(&cmd_line->options, 16, sizeof(parsed_option_t));
-				next_opt = rsh_blocks_vector_get_item(&cmd_line->options, cmd_line->options.size - 1, 16, parsed_option_t);
-
+				next_opt = new_option(cmd_line);
 				next_opt->buf[0] = '-', next_opt->buf[1] = ch, next_opt->buf[2] = '\0';
 				next_opt->name = next_opt->buf;
 				next_opt->parameter = NULL;
@@ -817,9 +832,6 @@ static void parse_cmdline_options(struct parsed_cmd_line_t* cmd_line)
 		}
 
 	} /* for */
-
-	cmd_line->n_files = n_files;
-	cmd_line->files = files;
 }
 
 /**
@@ -842,7 +854,7 @@ static void apply_cmdline_options(struct parsed_cmd_line_t *cmd_line)
 		}
 	}
 
-	/* if no formating options were specified at the command line */
+	/* if no formatting options were specified at the command line */
 	if (!opt.printf_str && !opt.template_file && !opt.sum_flags && !opt.fmt) {
 		/* copy the format from config */
 		opt.printf_str = conf_opt.printf_str;
@@ -876,10 +888,11 @@ static void apply_cmdline_options(struct parsed_cmd_line_t *cmd_line)
 
 	if (opt.embed_crc_delimiter == 0) opt.embed_crc_delimiter = conf_opt.embed_crc_delimiter;
 	if (!opt.path_separator) opt.path_separator = conf_opt.path_separator;
-	if (opt.find_max_depth < 0) opt.find_max_depth = conf_opt.find_max_depth;
-	if (!(opt.flags & OPT_RECURSIVE)) opt.find_max_depth = 0;
 	if (opt.flags & OPT_EMBED_CRC) opt.sum_flags |= RHASH_CRC32;
 	if (opt.openssl_mask == 0) opt.openssl_mask = conf_opt.openssl_mask;
+	if (opt.find_max_depth < 0) opt.find_max_depth = conf_opt.find_max_depth;
+	if (!(opt.flags & OPT_RECURSIVE)) opt.find_max_depth = 0;
+	opt.search_data->max_depth = opt.find_max_depth;
 
 	/* set defaults */
 	if (opt.embed_crc_delimiter == 0) opt.embed_crc_delimiter = " ";
@@ -966,7 +979,7 @@ void options_destroy(struct options_t* o)
  */
 static void make_final_options_checks(void)
 {
-	unsigned ff; /* formating flags */
+	unsigned ff; /* formatting flags */
 
 	if ((opt.flags & OPT_VERBOSE) && conf_opt.config_file) {
 		/* note that the first log_msg call shall be made after setup_output() */
@@ -983,12 +996,22 @@ static void make_final_options_checks(void)
 
 	ff = (opt.printf_str ? 1 : 0) | (opt.template_file ? 2 : 0) | (opt.fmt ? 4 : 0);
 	if ((opt.fmt & (opt.fmt - 1)) || (ff & (ff - 1))) {
-		die(_("too many formating options\n"));
+		die(_("too many formatting options\n"));
 	}
 
 	if (!opt.crc_accept) opt.crc_accept = file_mask_new_from_list(".sfv");
 
-	if (opt.openssl_mask) rhash_transmit(RMSG_SET_OPENSSL_MASK, 0, opt.openssl_mask, 0);
+	if (opt.openssl_mask) rhash_set_openssl_mask(opt.openssl_mask);
+}
+
+static struct parsed_cmd_line_t cmd_line;
+
+static void cmd_line_destroy(void)
+{
+	rsh_blocks_vector_destroy(&cmd_line.options);
+#ifdef _WIN32
+	LocalFree(cmd_line.warg);
+#endif
 }
 
 /**
@@ -998,9 +1021,8 @@ static void make_final_options_checks(void)
  */
 void read_options(int argc, char *argv[])
 {
-	struct parsed_cmd_line_t cmd_line;
-
 	opt.mem = rsh_vector_new_simple();
+	opt.search_data = file_search_data_new();
 	opt.find_max_depth = -1;
 
 	/* initialize cmd_line */
@@ -1008,6 +1030,7 @@ void read_options(int argc, char *argv[])
 	rsh_blocks_vector_init(&cmd_line.options);
 	cmd_line.argv = argv;
 	cmd_line.argc = argc;
+	rsh_install_exit_handler(cmd_line_destroy);
 
 	/* parse command line and apply encoding options */
 	parse_cmdline_options(&cmd_line);
@@ -1019,20 +1042,10 @@ void read_options(int argc, char *argv[])
 
 	apply_cmdline_options(&cmd_line); /* process the rest of command options */
 
-	/* options were processed, so we don't need them anymore */
-	rsh_blocks_vector_destroy(&cmd_line.options);
-
-	/* set the files and directories to be processed later */
-	opt.search_data = file_search_data_new(cmd_line.files, cmd_line.n_files, opt.find_max_depth);
-	opt.n_files = cmd_line.n_files;
-
-	free(cmd_line.files);
-
-#ifdef _WIN32
-	LocalFree(cmd_line.warg);
-#endif
+	/* options were processed, so we don't need them any more */
+	rsh_remove_exit_handler();
+	cmd_line_destroy();
 
 	make_final_options_checks();
-
 	set_default_sums_flags(argv[0]); /* detect default hashes from program name */
 }
